/**
 * Copyright (C) 2009 STMicroelectronics
 *
 * This file is part of "Mind Compiler" is free software: you can redistribute 
 * it and/or modify it under the terms of the GNU Lesser General Public License 
 * as published by the Free Software Foundation, either version 3 of the 
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact: mind-members@lists.minalogic.net
 *
 * Authors: Matthieu Leclercq
 * Contributors: 
 */

package org.objectweb.fractal.mind.idl;

import static org.objectweb.fractal.mind.BindingControllerImplHelper.checkItfName;
import static org.objectweb.fractal.mind.BindingControllerImplHelper.listFcHelper;
import static org.objectweb.fractal.mind.PathHelper.toAbsolute;

import java.util.HashMap;
import java.util.Map;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.error.ChainedErrorLocator;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.mind.PathHelper;
import org.objectweb.fractal.mind.idl.ast.IDL;
import org.objectweb.fractal.mind.idl.ast.Include;
import org.objectweb.fractal.mind.idl.ast.IncludeContainer;

public class BasicIncludeResolver implements IncludeResolver, BindingController {

  protected ThreadLocal<Map<String, IDL>> loadingIDLs = new ThreadLocal<Map<String, IDL>>();

  // ---------------------------------------------------------------------------
  // Client interfaces
  // ---------------------------------------------------------------------------

  /** The Loader interface used to load referenced IDLs. */
  public IDLLoader                        idlLoaderItf;

  // ---------------------------------------------------------------------------
  // Implementation of the IncludeResolver interface
  // ---------------------------------------------------------------------------

  public IDL resolve(final Include include,
      final IncludeContainer encapsulatingContainer,
      final Map<Object, Object> context) throws ADLException {

    Map<String, IDL> loadingIDLs = this.loadingIDLs.get();
    if (loadingIDLs == null) {
      loadingIDLs = new HashMap<String, IDL>();
      this.loadingIDLs.set(loadingIDLs);
    }

    final IDL idl = loadingIDLs.get(include.getPath());
    if (idl != null) {
      // The IDL to resolve is already loading (i.e. cycle in included files)
      return idl;
    }

    final IDL encapsulatingIDL;
    if (encapsulatingContainer instanceof IDL) {
      encapsulatingIDL = (IDL) encapsulatingContainer;
      assert !loadingIDLs.containsKey(encapsulatingIDL.getName());
      loadingIDLs.put(encapsulatingIDL.getName(), encapsulatingIDL);
    } else {
      encapsulatingIDL = null;
    }

    String path = include.getPath();
    if (encapsulatingIDL != null) {
      final String encapsulatingIDLName = encapsulatingIDL.getName();
      final String dirName;
      if (encapsulatingIDLName.startsWith("/")) {
        dirName = PathHelper.getParent(encapsulatingIDLName);
      } else {
        dirName = PathHelper.fullyQualifiedNameToDirName(encapsulatingIDLName);
      }
      path = toAbsolute(dirName, path);
      // set absolute path in node without first '/'
      include.setPath(path.substring(1));
    }

    try {
      return idlLoaderItf.load(path, context);
    } catch (final ADLException e) {
      ChainedErrorLocator.chainLocator(e, include);
      throw e;
    } finally {
      if (encapsulatingIDL != null) {
        loadingIDLs.remove(encapsulatingIDL.getName());
      }
    }
  }

  // ---------------------------------------------------------------------------
  // Overridden BindingController methods
  // ---------------------------------------------------------------------------

  public void bindFc(final String itfName, final Object value)
      throws NoSuchInterfaceException, IllegalBindingException {
    checkItfName(itfName);

    if (itfName.equals(IDLLoader.ITF_NAME)) {
      idlLoaderItf = (IDLLoader) value;
    } else {
      throw new NoSuchInterfaceException("No client interface named '"
          + itfName + "'");
    }

  }

  public String[] listFc() {
    return listFcHelper(IDLLoader.ITF_NAME);
  }

  public Object lookupFc(final String itfName) throws NoSuchInterfaceException {
    checkItfName(itfName);

    if (itfName.equals(IDLLoader.ITF_NAME)) {
      return idlLoaderItf;
    } else {
      throw new NoSuchInterfaceException("No client interface named '"
          + itfName + "'");
    }
  }

  public void unbindFc(final String itfName) throws NoSuchInterfaceException,
      IllegalBindingException {
    checkItfName(itfName);

    if (itfName.equals(IDLLoader.ITF_NAME)) {
      idlLoaderItf = null;
    } else {
      throw new NoSuchInterfaceException("No client interface named '"
          + itfName + "'");
    }
  }
}
